import logging
import types
from inspect import isclass

LOG = logging.getLogger(__name__)


def copy_func(f, name=None):
    return types.FunctionType(f.func_code, f.func_globals, name or f.func_name,
        f.func_defaults, f.func_closure)


class Registry(object):

    def __init__(self):
        self._registry = {}
        self._cache = []
        self._aliases = {}

    def get_commands(self, key):
        try:
            commands = self._registry[key]
        except KeyError:
            commands = {}
        return commands

    def is_cached(self, func):
        return func in self._cache

    def _register(self, key, func, name):
        LOG.debug('registering "%s" to "%s"', name, key)
        commands = self.get_commands(key)
        if name in commands.keys():
            raise ValueError("{0} already registered to {1}".format(name, key))
        if isclass(func):
            commands[name] = func
        else:
            # we copy the function so alias will show the proper usage line with the alias name
            commands[name] = copy_func(func, name)
        self._registry[key] = commands

    def register(self, func, name, key=None, aliases=[]):
        if func not in self._cache:
            LOG.debug('caching command "%s"', name)
            self._cache.append(func)
        if any(key):
            self._register(key, func, name)
            for alias in aliases:
                self._register(key, func, alias)
            for alias in [a for a, f in self._aliases.items() if f is func]:
                self._register(key, func, alias)
        else:
            for alias in aliases:
                LOG.debug('caching alias "%s" for command "%s"', alias, name)
                self._aliases[alias] = func

    def bind(self, func, cli, handler=None, aliases=[]):
        if handler:
            key = (cli, handler)
        else:
            key = (cli,)
        self.register(func, func.__name__, key, aliases)


registry = Registry()


class HandlerRegistrationMixin(type):
    def __new__(cls, clsname, bases, attrs):
        newclass = super(HandlerRegistrationMixin, cls).__new__(cls, clsname, bases, attrs)
        if hasattr(newclass.State, 'cli') and newclass.State.cli is not None and newclass.State.cli != "self":
            registry.register(newclass, clsname, (newclass.State.cli,))
            for k, v in attrs.items():
                if registry.is_cached(v):
                    registry.register(v, k, (newclass.State.cli, clsname))
        return newclass


class CLIRegistrationMixin(type):
    def __new__(cls, clsname, bases, attrs):
        newclass = super(CLIRegistrationMixin, cls).__new__(cls, clsname, bases, attrs)
        for k, v in attrs.items():
            if registry.is_cached(v):
                registry.register(v, k, (clsname,))
        return newclass


def command(*args, **kwargs):
    """
    Decorator that can expose a function as a command for a cli or handler
    """
    invoked = bool(not args or kwargs)
    if not invoked:
        func, args = args[0], ()

    def register(func):
        name = func.__name__
        cli = kwargs.get('cli', None)
        handler = kwargs.get('handler', None)
        if handler:
            key = (cli, handler)
        else:
            key = (cli,)
        aliases = kwargs.get('aliases', [])
        if not isinstance(aliases, list):
            raise ValueError('command decorator aliases takes a list, ' \
                             'just use "alias" for a single string alias!')
        alias = kwargs.get('alias', [])
        if isinstance(alias, list):
            aliases += alias
        else:
            aliases += [alias]
        registry.register(func, name, key, aliases)
        return func
    return register if invoked else register(func)



from pyul.coreUtils import DotifyDict


class State(DotifyDict):
    """
    A class to provide a way to combine
     - state settings | by the programmer
     - config settings | by the users environment
     - options settings | by the user at command runtime
    To produce a final "state" of the configuration
    """

    def __init__(self):
        self.cli = None
        self.state_list = list()
        self.options_list = list()
        self.config_list = list()

    def add_options(self, options):
        self.options_list.append(options)

    def add_state(self, state):
        self.state_list.append(state)

    def add_config(self, config):
        self.config_list.append(config)

    def compile(self):
        for state in reversed(self.state_list):
            self.update(state)

        for config in self.config_list:
            self.update(config)

        for option in self.options_list:
            self.update(option)

        self.pop('state_list')
        self.pop('config_list')
        self.pop('options_list')
        self.cli.state = self
        for handler in self.cli.handlers:
            handler.state = self

state = State()


class StateMixin(object):
    """
    Mixin that provides static configuration settings to instance
    objects. State settings cannot start with a _.
    """

    def __init__(self, *args, **kwargs):
        # Get a List of all the Classes we in our MRO, find any attribute named
        #     State on them, and then merge them together in order of MRO
        states = reversed([x.State for x in self.__class__.mro()
                          if hasattr(x, "State")])
        final_state = DotifyDict()

        # Merge the State classes into one dict
        for state in states:
            final_state.update(DotifyDict(dict([x for x in state.__dict__.items()
                                           if not x[0].startswith("_")])))

        # Update the final state with any kwargs passed in
        for key in final_state.keys():
            if key in kwargs:
                final_state[key] = kwargs.pop(key)

        self._state = final_state

        super(StateMixin, self).__init__()



from inspect import getargspec


class NoSuchCommand(ValueError):
    def __init__(self, command, supercommand):
        super(NoSuchCommand, self).__init__("No such command: %s" % command)
        self.command = command
        self.supercommand = supercommand


class CommandMissingDefaults(ValueError):
    def __init__(self, command):
        spec = getargspec(command)
        msg = """
        command '{0}' does not have defaults for every argument it takes
        ARGS: {1}
        DEFAULTS: {2}
        """
        super(CommandMissingDefaults, self).__init__(msg.format(command.__name__,
                                                                spec.args[1:],
                                                                spec.defaults))



import logging

logging_setup = False

SIMPLE = "[%(levelname)s] %(message)s"
VERBOSE = "%(levelname)-6s %(asctime)s line:%(levelno)-3d %(name)-25s  : %(message)s"


def enable_logging(fmt=SIMPLE, level=logging.DEBUG):
    global logging_setup

    if not logging_setup:
        logging.basicConfig(format=fmt,
                            datefmt="%d-%m %H:%M:%S",
                            level=level)
        logging_setup = True



from battalion.api import *


class myCLI(CLI):    
    """
    Toplevel program - myCLI
    """
    class State:
        version = '0.0.1'
    
    def normal_function(self, val):
        return "Josh"

    @command
    def command_output(cli, val='test'):
        """"""
        return "Boom"

    @command(alias="hello_world")
    def world(cli, name="Hello"):
        """
        Prints "{name} World!"
        """
        print "{0} World!".format(name)
        print cli

    @command
    def add(cli, num1, num2, mul=2):
        """
        Add {num1} to {num2}
        """
        return (float(num1) + float(num2)) * float(mul)


class myhandler(Handler):
    """
    Organizational container of commands which can also add state variables
    """
    class State:
        version = '0.0.2'  # Handlers can have their own versioning
        cli = 'myCLI'
        static_var = "Kyle"

    @command
    def hello(cli):
        """
        Prints "Hello, World!"
        """
        name = "Kyle"
        name = cli.normal_function(cli.state.static_var)
        name = cli.command_output(val=name)
        cli.myhandler.greeting(name=name)

    @command(alias='add2')
    def add(cli, num1, num2):
        """
        Add {num1} to {num2}
        """
        return float(num1) + float(num2)


@command(cli='myCLI', handler='myhandler', alias='greet') # alias or aliases are great ways to bridge gaps during refactoring of api
@command(cli='myCLI') # multiple command decorators can be applied to place a command into multiple areas for ease of use but still benefits from a single source of code
                      # NOTE: state will differ since its being built up with handler or without the handler
def greeting(cli, greeting="Hello", name="World"):
    """
    Prints "{greeting}, {name}"
    """
    print "{0}, {1}!".format(greeting, name)
    print cli.state.static_var

@command(cli='myCLI')
def test(cli, name="DAMN"):
    """
    Prints {name}!
    """
    print name
    
registry.bind(test, 'myCLI', 'myhandler')

if __name__ == "__main__":
    myCLI.main()


from battalion.api import *

class myCLI(CLI):    
    """
    Toplevel program - myCLI
    """
    class State:
        version = '0.0.1'

    @command
    def hello(cli, name="Hello"):
        """
        Prints "{name} World!"
        """
        cli.log.info("{0} World!".format(name))
        

class myhandler(Handler):
    """
    Organizational container of commands which can also add state variables
    """
    class State:
        version = '0.0.2'  # Handlers can have their own versioning
        cli = 'myCLI'
        static_var = "Kyle"

    @command
    def hello(cli):
        """
        Prints "Hello, World!"
        """
        cli.log.info("Hello")


if __name__ == "__main__":
    myCLI.main()


import os
import sys
import re
import logging
import six
import yaml
from functools import partial
from inspect import getdoc, cleandoc, isclass, getargspec, getcallargs
from pyul.coreUtils import DotifyDict
from docopt import docopt

from .exceptions import NoSuchCommand, CommandMissingDefaults
from .registry import CLIRegistrationMixin, HandlerRegistrationMixin, registry
from .state import StateMixin, state
from .log import enable_logging

LOG = logging.getLogger(__name__)


def cleanup_data(data):
    new_data = {}
    for k, v in data.items():
        if k.startswith('--'):
            k = k[2:]
        if k.startswith('-'):
            k = k[1:]
        k = k.replace('-', '_')
        if k in ['help', 'version', 'cli', 'options', 'column_padding', 'default_config', 'config']:
            continue
        new_data[k] = v
    return DotifyDict(new_data)


def get_command_spec(command):
    spec = getargspec(command)
    positional = [None] * (len(spec.args) - len(spec.defaults or []))
    kwargs = getcallargs(command, *positional)
    kwargs.pop('cli')
    return kwargs


def parse_doc_section(name, source):
    pattern = re.compile('^([^\n]*' + name + '[^\n]*\n?(?:[ \t].*?(?:\n|$))*)',
                         re.IGNORECASE | re.MULTILINE)
    return [s.strip() for s in pattern.findall(source)]


class BaseCommand(StateMixin):
    """
    Base class for battalion commandline app
    """

    class State:
        cli = None
        version = 'UNKNOWN'
        default_config = None
        config_file = None
        options = [('-h, --help', 'Show this screen.'),
                   ('--version', 'Show version.')]

    @property
    def name(self):
        return self.__class__.__name__

    @property
    def handler(self):
        return None

    @property
    def cli(self):
        return self._state.cli

    @property
    def key(self):
        return (self.cli, self.name)

    @property
    def commands(self):
        return registry.get_commands(self.key)

    @property
    def handlers(self):
        return [h for h in self.commands if isinstance(h, Handler)]

    @property
    def docstring(self):
        raise NotImplementedError

    @property
    def docopt_options(self):
        return {'options_first': True,
                'version': self._state.version}

    def format_command_args(self, command, kwargs):
        new_kwargs = {}
        command_kwargs = get_command_spec(command)
        for k, v in sorted(kwargs.items()):
            if k.startswith('--'):
                k = k[2:]
            if k.startswith('-'):
                k = k[1:]
            k = k.replace('-', '_')
            k = k.replace('<', '')
            k = k.replace('>', '')
            if v is None:
                v = new_kwargs[k] or command_kwargs[k] or None
            if k in command_kwargs:
                new_kwargs[k] = v
        return new_kwargs

    def get_options(self, argv):
        options = docopt(self.docstring,
                         argv,
                         **self.docopt_options)
        return options

    def get_command(self, options):
        command_name, args = options.pop('<command>'), options.pop('<args>')
        if command_name is None or command_name is False:
            raise SystemExit(self.docstring)
        try:
            command = self.commands[command_name]
        except KeyError:
            raise NoSuchCommand(command_name, self)
        state.add_state(cleanup_data(self._state))
        state.add_options(cleanup_data(options))
        return command, args

    def dispatch(self, argv):
        options = self.get_options(argv)
        command, args = self.get_command(options)
        return self.run(command, args)

    def run(self, command, args):
        command_options = docopt(cleandoc(command.__autodoc__), args)
        kwargs = self.format_command_args(command, command_options)
        state.add_options(kwargs)
        state.compile()

        if state.debug:
            LOG.debug("State:\n{0}".format(state))
        return command(state.cli, **kwargs)


class AutoDocCommand(BaseCommand):
    """
    Class that supports generating the docopt docstrings for the class and
    it's command functions.
    """
    class State:
        column_padding = 30

    def __init__(self):
        if self.__doc__ is None:
            self.__doc__ = ""
        super(AutoDocCommand, self).__init__()
        # We check if autodoc has already happend so that
        # so that test frameworks can keep generating new
        # instances of the same class without redocumenting
        # which will cause a failure
        if not hasattr(self, "__autodoc__"):
            self.generate_class_doc()
            self.generate_commands_doc()

    @property
    def docstring(self):
        return cleandoc(self.__autodoc__)

    def generate_class_doc(self):
        LOG.debug('Documenting %s', self.name)
        new_doc = getdoc(self) or """{0}""".format(self.name)
        new_doc += "\n\n"
        new_doc += self.generate_usage()
        new_doc += self.generate_options()
        new_doc += self.generate_commands()
        self.__autodoc__ = cleandoc(new_doc)

    def generate_commands_doc(self):
        for name, func in self.commands.items():
            if isclass(func) and issubclass(func, Handler):
                LOG.debug('Documenting Command %s', name)
                self.commands[name] = func()
            else:
                LOG.debug('Documenting Command %s', name)
                new_command_doc = getdoc(func) or """{0}""".format(name)
                new_command_doc += "\n\n"
                new_command_doc += self.generate_command_usage(name, func)
                new_command_doc += self.generate_command_options(func)
                func.__autodoc__ = cleandoc(new_command_doc)
                self.commands[name] = func

    def generate_usage(self):
        docstring = ""
        if "Usage:" not in self.__doc__:
            docstring += "Usage:\n"
            docstring += "    {0} [options] <command> [<args>...]\n".format(self.name)
            docstring += "    {0} [options]\n\n".format(self.name)
        return docstring

    def generate_options(self):
        if "Options:" not in self.__doc__:
            docstring = "Options:\n"
            for flags, desc in self._state.options:
                if flags == '--config=<CONFIG>' and flags not in docstring:
                    if isinstance(self, CLI):
                        if self._state.default_config is None:
                            self._state.default_config = '~/.{name}/{name}.yaml'.format(name=self.name)
                        desc = desc.format(self._state.default_config)
                    else:
                        continue
                docstring += "    {0:<{2}} {1}\n".format(flags,
                                                         desc,
                                                         self._state.column_padding)
            docstring += "\n"
        return docstring

    def generate_commands(self):
        if "Commands:" not in self.__doc__:
            docstring = "Commands:\n"
            for k, v in self.commands.items():
                docstring += "    {0:<{2}} {1}\n".format(k,
                                                         getdoc(v),
                                                         self._state.column_padding)
            docstring += "\n"
        return docstring

    def generate_command_usage(self, name, command):
        docstring = ""
        if "Usage:" not in command.__doc__:
            docstring += "Usage:\n    {0} [options]\n".format(name)
            args = getargspec(command).args
            spec = get_command_spec(command)
            if 'cli' in args:
                args = args[1:]
            if args:
                docstring += "    {0} ".format(name)
                for arg_name in args:
                    if spec[arg_name] is not None:
                        docstring += "[<{0}>] ".format(arg_name)
                    else:
                        docstring += "<{0}> ".format(arg_name)
                docstring += "\n"
            docstring += "\n"
        return docstring

    def generate_command_options(self, command):
        docstring = ""
        if "Options:" not in command.__doc__:
            args = get_command_spec(command)
            if args:
                docstring += "Options:\n"
                for arg, default in args.items():
                    flag_def = "--{0}=<{1}>".format(arg,
                                                    arg.upper())
                    docstring += "    {0:<{3}} {1} [default: {2}]\n".format(flag_def,
                                                                            ' ',
                                                                            default,
                                                                            self._state.column_padding)
                docstring += "\n"
        return docstring


@six.add_metaclass(HandlerRegistrationMixin)
class Handler(AutoDocCommand):

    def __init__(self):
        super(Handler, self).__init__()

    def __getattr__(self, name):
        if name in self.commands:
            return partial(self.commands[name], state.cli)
        raise AttributeError


@six.add_metaclass(CLIRegistrationMixin)
class CLI(AutoDocCommand):
    class State:
        options = [('-d, --debug', 'Show debug messages'),
                   ('--config=<CONFIG>', 'The config filepath [default: {0}]')]

    @classmethod
    def main(cls, argv=None):
        return cls()(argv)

    def __init__(self):
        self.log = logging.getLogger(self.name)
        state.cli = self
        super(CLI, self).__init__()

    def __call__(self, argv=None):
        if argv is None:
            argv = sys.argv[1:]
        try:
            rv = None
            self.setup_logging()
            rv = self.dispatch(argv=argv)
        except KeyboardInterrupt:
            LOG.error("\nAborting.")
            sys.exit(1)
        except NoSuchCommand as e:
            LOG.error("No such command: %s", e.command)
            LOG.error("")
            LOG.error("\n".join(parse_doc_section("commands:", getdoc(e.supercommand))))
            sys.exit(1)
        finally:
            if rv:
                print rv

    def __getattr__(self, name):
        if name in self.commands:
            cmd = self.commands[name]
            if isinstance(cmd, Handler):
                return cmd
            else:
                return partial(cmd, self)
        raise AttributeError

    @property
    def key(self):
        return (self.name,)

    def setup_logging(self):
        enable_logging(level=logging.INFO)

    def dispatch(self, argv):
        options = self.get_options(argv)
        self.load_config(options)
        command, args = self.get_command(options)
        if isinstance(command, Handler):
            return command.dispatch(args)
        else:
            args.insert(0, command.func_name)
            return super(CLI, self).dispatch(argv)

    def load_config(self, options):
        config_filepath = os.path.expanduser(options['--config'])
        self._state.config_file = config_filepath
        if os.path.exists(config_filepath):
            with open(self._state.config_file, 'r') as ymlfile:
                config = DotifyDict(data=yaml.load(ymlfile))
                state.add_config(cleanup_data(config))
